from copy import copy
#00 01 10 11

sign_table_ = [
    [0,0,0,1],   #|
    [0,1,1,1],   #&
    [1,1,0,0],   #-
    [1,0,0,1]    #^
]

op_table_ = [
    [lambda a,b: a&b, lambda a,b: b-a,  lambda a,b: a-b, lambda a,b: a|b,],   #|
    [lambda a,b: a|b, lambda a,b: a-b,  lambda a,b: b-a, lambda a,b: a&b,],   #&
    [lambda a,b: b-a, lambda a,b: a&b,  lambda a,b: a|b, lambda a,b: a-b,],   #-
    [lambda a,b: a^b, lambda a,b: a^b,  lambda a,b: a^b, lambda a,b: a^b,],   #^
    [lambda a,b: b<a, lambda a,b: len(a&b)==0, lambda a,b: False, lambda a,b: a<b],    #<
    [lambda a,b: b<=a, lambda a,b: len(a&b)==0, lambda a,b: False, lambda a,b: a<=b],    #<=
]

iop_table_ = [
    [lambda a,b: a.__iand__(b), lambda a,b: b.__isub__(a),  lambda a,b: a.__isub__(b), lambda a,b: a.__ior__(b),],  #|=
    [lambda a,b: a.__ior__(b), lambda a,b: a.__isub__(b),  lambda a,b: b.__isub__(a), lambda a,b: a.__iand__(b),],  #&=
    [lambda a,b: b.__isub__(a), lambda a,b: a.__iand__(b),  lambda a,b: a.__ior__(b), lambda a,b: a.__isub__(b),],  #-=
    [lambda a,b: a.__ixor__(b), lambda a,b: a.__ixor__(b),  lambda a,b: a.__ixor__(b), lambda a,b: a.__ixor__(b),]  #^=
]

def _aux1(table):
    return [[arr[0:2], arr[2:4]] for arr in table]

sign_table = _aux1(sign_table_)
op_table = _aux1(op_table_)
iop_table = _aux1(iop_table_)

def create_signset(set_cls):
    def _op_gen(tb_idx):
        def f(self, x):
            funcs, signs = op_table[tb_idx], sign_table[tb_idx]
            a,b = self._data,x._data
            m,n = self.sign,x.sign
            ret = SignSet(sign = signs[m][n])
            ret._data = funcs[m][n](a,b)
            return ret            
        return f

    def _iop_gen(tb_idx):
        def f(self, x):
            funcs, signs = iop_table[tb_idx], sign_table[tb_idx]
            a,b = self._data,x._data
            m,n = self.sign,x.sign
            funcs[m][n](a,b)
            self.sign = signs[m][n]
            return self
        return f

    def _cmp_op_gen(tb_idx):
        def f(self, x):
            funcs = op_table[tb_idx]
            a,b = self._data,x._data
            m,n = self.sign,x.sign
            return funcs[m][n](a,b)
        return f  
                        
    _op_impls = list(map(_op_gen, range(4)))
    _iop_impls = list(map(_iop_gen, range(4)))
    _cmp_op_impls = list(map(_cmp_op_gen, (4,5)))

    class SignSet:
        def __init__(self, *args ,sign = 1):
            self.sign = sign
            self._data = set_cls(*args)

        def __or__(self,x):
            return _op_impls[0](self,x)

        def __and__(self,x):
            return _op_impls[1](self,x)

        def __sub__(self,x):
            return _op_impls[2](self,x)

        def __xor__(self,x):
            return _op_impls[3](self,x)

        def __ior__(self,x):
            return _iop_impls[0](self,x)

        def __iand__(self,x):
            return _iop_impls[1](self,x)

        def __isub__(self,x):
            return _iop_impls[2](self,x)

        def __ixor__(self,x):
            return _iop_impls[3](self,x)

        def __lt__(self,x):
            return _cmp_op_impls[0](self,x)

        def __le__(self,x):
            return _cmp_op_impls[1](self,x)

        def __gt__(self,x):
            return x<self

        def __ge__(self,x):
            return x>=self

        def __eq__(self,x):
            return self.sign == x and self._data == x._data
        
        def __neq__(self,x):
            return not (self == x)

        def __neg__(self):
            ret = self.copy()
            ret.neg_self()
            return ret

        def __contains__(self,x):
            return self.sign == (x in self._data)

        def copy(self):
            return copy(self)

        def neg_self(self):
            self.sign = not self.sign

        def is_empty(self):
            return (not self.sign) or (len(self._data) == 0)

        def clear(self):
            self.sign = 1
            self._data.clear()

        def add(self,x):
            return self._data.add(x)

        def remove(self,x):
            return self._data.remove(x)

        def discard(self, x):
            return self._data.discard(x)

        def to_str(self, ts = str):
            s = ts(self._data)
            if not self.sign: s = '-'+s
            return s

        def __str__(self):
            return self.to_str()

        def __repr__(self):
            return self.to_str(repr)

    return SignSet


import asyncio
import logging
from collections import deque
from contextlib import contextmanager, asynccontextmanager

__all__ = ['CoroDeque', 'CoroEvent']

class CoroEvent(asyncio.Event):
    def __init__(self):
        self.wait_num = 0
        super().__init__()

    async def wait(self):
        self.wait_num += 1
        await super().wait()
        self.wait_num -= 1


class CoroDeque:
    class NoWait:
        pass

    def __init__(self, iterable=None, maxlen=None):
        kwargs = {}
        if iterable != None:
            kwargs['iterable'] = iterable
        if maxlen != None:
            kwargs['maxlen'] = maxlen
        self._q = deque(**kwargs)
        self.loop = asyncio.get_event_loop()

        self._pop_event = asyncio.Event()
        self._put_event = asyncio.Event()
        self._pop_lock = asyncio.Lock()
        self._put_lock = asyncio.Lock()
        self.pop_wait_num = 0
        self.put_wait_num = 0

        # if self.is_full():
        #     self._put_lock.acquire()
        # if len(self) == 0:
        #     self._pop_lock.acquire()

    @property
    def maxlen(self):
        return self._q.maxlen

    def __len__(self):
        return len(self._q)

    def append_nw(self, item):
        self._q.append(item)
        self._after_put()

    async def append(self, item):
        async with self._when_put():
            self._q.append(item)

    def appendleft_nw(self, item):
        self._q.appendleft(item)
        self._after_put()

    async def appendleft(self, item):
        async with self._when_put():
            self._q.appendleft(item)

    def pop_nw(self):
        res = self._q.pop()
        self._after_pop()
        return res

    async def pop(self):
        async with self._when_pop():
            res = self._q.pop()
        return res

    def popleft_nw(self):
        res = self._q.popleft()
        self._after_pop()
        return res

    async def popleft(self):
        async with self._when_pop():
            res = self._q.popleft()
        return res

    def put_nw(self, item):
        self.append_nw(item)

    async def put(self, item):
        await self.append(item)

    def popout_nw(self):
        return self.popleft_nw()

    async def popout(self):
        return await self.popleft()

    async def wait(self):
        """
        等待不为空
        """
        while len(self) == 0:
            async with self._acquire_pop_lock():  # 获取并释放
                pass

    async def wait_pop(self):
        await self._pop_event.wait()

    async def wait_done(self):
        while len(self) != 0:
            await self._pop_event.wait()

    def wait_b(self):
        if len(self) == 0:
            self.loop.run_until_complete(
                self.wait())

    async def wait_not_full(self):
        while self.is_full():
            async with self._acquire_put_lock():  # 获取并释放
                pass

    def wait_not_full_b(self):
        if self.is_full():
            self.loop.run_until_complete(
                self.wait_not_full())

    async def push_no_waits(self):
        '''
        放入停止等待信息，直到当前所有等待取出数据的对象不再等待
        '''
        NoWait = self.NoWait
        wait_num = self.pop_wait_num
        put_num = wait_num - len(self)
        if put_num > 0:
            for _ in range(put_num):
                await self.put(NoWait)

    def is_full(self):
        maxlen = self.maxlen
        return self.maxlen is not None and len(self) >= maxlen

    def clear(self):
        self._q.clear()
        self._after_pop()

    def __str__(self):
        return self.to_str(str)

    def __repr__(self):
        return self.to_str(repr)

    def to_str(self, f=str):
        return f'AsyncDeque{f(self._q)[5:]}'

    def __iter__(self):
        return iter(self._q)

    def _after_pop(self):
        self._set_event(self._pop_event)
        self._pop_event = asyncio.Event()
        self._release_lock(self._put_lock)

    @asynccontextmanager
    async def _acquire_pop_lock(self):
        try:
            while len(self) == 0:
                self.pop_wait_num += 1
                await self._pop_lock.acquire()
                self.pop_wait_num -= 1
            yield
        finally:
            if len(self) > 0:  # not empty
                self._release_lock(self._pop_lock)

    @asynccontextmanager
    async def _when_pop(self):
        # before pop
        try:
            if len(self) == 0:
                async with self._acquire_pop_lock():
                    yield  # pop
            else:
                yield  # pop
        # after pop
        finally:
            self._after_pop()

    def _after_put(self):
        self._set_event(self._put_event)
        self._put_event = asyncio.Event()
        self._release_lock(self._pop_lock)

    @asynccontextmanager
    async def _acquire_put_lock(self):
        try:
            while self.is_full():
                self.put_wait_num += 1
                await self._put_lock.acquire()
                self.put_wait_num -= 1
            yield  # put
        finally:
            if not self.is_full():
                self._release_lock(self._put_lock)

    @asynccontextmanager
    async def _when_put(self):
        try:
            # before put
            if self.is_full():
                async with self._acquire_put_lock():
                    yield  # put
            else:
                yield  # put
        finally:
            # after put
            self._after_put()

    @staticmethod
    def _set_event(*events):
        for event in events:
            if not event.is_set():
                event.set()

    @staticmethod
    def _clear_event(*events):
        for event in events:
            if event.is_set():
                event.clear()

    @staticmethod
    def _release_lock(*locks):
        for lock in locks:
            if lock.locked():
                lock.release()



import uuid
import base64
import base58

__all__ = [
    'to_uuid_b64url_str',
    'make_uuid4_b64url_str',
    'make_uuid4_b58_str'
]


def to_uuid_b64url_str(uuid):
    return base64.urlsafe_b64encode(uuid.bytes).decode(encoding='utf8')

def to_uuid_b58url_str(uuid):
    return base58.b58encode(uuid.bytes).decode(encoding='utf8')


def make_uuid4_b64url_str():
    return to_uuid_b64url_str(uuid.uuid4())

def make_uuid4_b58_str():
    return to_uuid_b58url_str(uuid.uuid4())






# import requests
# import contextlib
# from .path import Path
# from urllib.parse import urlparse
# from tqdm import tqdm
# from .pbar import file_proc_bar

# def download_bar(iterable, chunk_size = None, total_size = None, exist_size = 0):
#     def bar():
#         with file_proc_bar(total=total_size) as pbar:
#             pbar.update(exist_size)
#             for x in iterable:
#                 yield x
#                 pbar.update(chunk_size)
#     return bar()

# class URL:
#     def __init__(self, url):
#         self.url = url
#         self.o = urlparse(url)

#     def to_str(self, ts = str):
#         return ts(self.url)

#     def __str__(self):
#         return self.to_str(str)

#     def __repr__(self):
#         return self.to_str(repr)

#     @property
#     def path(self):
#         return self.o.path
    
#     @property
#     def name(self):
#         return self.path.split('/')[-1]

#     def download(self, save_path, continuous = False, enable_print = True, enable_bar = True, chunk_size = 4<<10):
#         url = self.url
#         save_path = Path(save_path)
#         r = requests.get(url, stream = True)
#         total_size = int(r.headers['Content-Length'] )
#         exist_size = 0

#         if continuous and save_path.exists():
#             assert(save_path.is_file())
#             exist_size = save_path.size
#             if exist_size:
#                 r.close()
#                 if exist_size == total_size:    #has downloaded
#                     if enable_print:
#                         print(f'Downloading: {url}')
#                         print(f'Save as: {save_path}')
#                         if enable_bar:
#                             with file_proc_bar(total=total_size) as pbar:
#                                 pbar.update(total_size)
#                     return

#                 #redo request
#                 r = requests.get(url, stream = True, headers = {'Range': f'bytes={exist_size}-'})
#                 fw = save_path.open('ab')
#             else:   #none was downloaded
#                 fw = save_path.open('wb')
#         else:
#             save_path.prnt.mkdir()
#             fw = save_path.open('wb')

#         it = r.iter_content(chunk_size=chunk_size)
#         if enable_print: 
#             if enable_bar: 
#                 it = download_bar(it, chunk_size = chunk_size, total_size = total_size, exist_size = exist_size)
#             print(f'Downloading: {url}')
#             print(f'Save as: {save_path}')
#         for data in it:
#             fw.write(data)
#         fw.close()
#         r.close()
            

# __all__ = ['URL']


import re

from .iterutil import *
from .path import *
from .adaptor import *
from .random import *
del random
# from .util import *
# from .url import *
# from .pbar import *
# from .printutil import *

__all__ = list(filter(lambda s: not(re.match('_+',s) or re.search('_+$',s)), globals() ))


#-*-conding:utf-8-*-
import os
import pathlib
import shutil
from distutils import file_util, dir_util
from zipfile import ZipFile

# from .pbar import file_proc_bar
from .util import get_unique_name
# from itertools import chain

from cbutil.util import dfs

_Path = type(pathlib.Path(''))

__all__ = ['Path']


class Path(_Path):
    _Path = _Path

    def __init__(self, *args, **kwargs):
        pass

    def __new__(cls, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0:
            p = args[0]
            if type(p) == Path:
                return p
        return super().__new__(cls, *args, **kwargs)

    @staticmethod
    def setcwd(path):
        os.chdir(path)
    
    @staticmethod
    def getcwd():
        return os.getcwd()

    @property
    def prnt(self):
        return Path(super().parent)

    @property
    def ext(self):
        return super().suffix[1:]

    @property
    def size(self):
        return self.stat().st_size
    
    @property
    def rsize(self):
        '''real size'''
        return self.stat().st_rsize

#begin iter

    def get_son_iter(self, *filters):
        if self.is_dir():
            if len(filters) == 0:
                return super().iterdir()
            return filter(lambda x: all(map(lambda f: f(x), filters)), super().iterdir())
        else:
            return iter([])

    def get_file_son_iter(self, *filters):
        return self.get_son_iter(Path.is_file, *filters)

    def get_dir_son_iter(self, *filters):
        return self.get_son_iter(Path.is_dir, *filters)

    def get_sibling_iter(self, *filters):
        return self.prnt.get_son_iter(lambda x: x!=self, *filters)

    @property
    def son_iter(self):
        return self.get_son_iter()

    @property
    def sibling_iter(self, *filters):
        return self.get_sibling_iter()

    @property
    def file_son_iter(self):
        return self.get_file_son_iter()

    @property
    def dir_son_iter(self):
        return self.get_dir_son_iter()

    @property
    def sons(self):
        return list(self.son_iter)
    
    @property
    def siblings(self):
        return list(self.sibling_iter)

    @property
    def file_sons(self):
        return list(self.file_son_iter)

    @property
    def dir_sons(self):
        return list(self.dir_son_iter)
#end iter


    @property
    def str(self):
        return self.__str__()

    @property
    def quote(self):
        if Path._Path == pathlib.WindowsPath:
            s = self.str.replace('"', r'\"')
            return f'"{s}"'

    def is_empty_dir(self):
        if self.is_dir():
            for _ in self.son_iter:
                return False
            return True
        return False

    def rel_to(self,path):
        return super().relative_to(path)

    def open(self, mode, buffering=-1, encoding='utf8', newline='', *args, **kwargs):
        if 'b' in mode:
            encoding=None
            newline=None
        if not self.prnt.exists():
            self.prnt.mkdir()
        return super().open(mode=mode,buffering=buffering, encoding=encoding, newline=newline, *args,**kwargs)
    
    def write(self, content):
        if isinstance(content, bytes):
            with self.open('wb') as f:
                f.write(content)
        elif isinstance(content, str):
            with self.open('w') as f:
                f.write(content)
        else:
            raise TypeError
        
    def read_text(self, encoding='utf8'):
        with self.open('r',encoding=encoding) as f:
            return f.read()
    
    def mkdir(self, *args, update =False, parents =True, **kwargs):
        '''
        默认文件夹不存在则忽略，会自动创建不存在的父路径
        '''
        if self.exists():
            if update:
                self.remove()
                super().mkdir(*args, parents=parents, **kwargs)
        else:
            super().mkdir(*args, parents=parents, **kwargs)

    def remove(self):
        if self.exists():
            if self.is_dir():
                shutil.rmtree(self.to_str())
            else:
                os.remove(self.to_str())

    def remove_sons(self):
        assert(self.is_dir())
        for son in self.son_iter:
            son.remove()
    
    def to_str(self):
        return str(self)

    def copy_to(self, dst:'Path', is_prefix = True, update = True, ignore_exist = True):
        '''
        会自动创建当前不存在的父目录
        
        若is_prefix == False:
        将该路径复制，复制后的路径为dst
        若is_prefix == True:
        则将该路径复制到dst目录下，复制后的名字不变
        
        若路径为目录，则会递归复制，
        '''
        if is_prefix:
            dst = dst/self.name
        
        if self.is_file():
            if not dst.prnt.exists():
                dst.prnt.mkdir()
            file_util.copy_file(self.str, dst.str, update)
        else:
            if not dst.prnt.exists():
                dst.prnt.mkdir()
            dir_util.copy_tree(self.str, dst.str, update)


    def copy_sons_to(self, dst, update = True):
        dst = Path(dst)
        assert self.is_dir()
        for son in self.son_iter:
            son.copy_to(dst/son.name, is_prefix=False, update=update)

    def make_copy(self, name:str, update=False):
        self.copy_to(self.prnt/name, is_prefix=False, update=update)

    def move_to(self, dst, is_prefix=True, update=True, recursive=True):
        dst = Path(dst)
        if is_prefix:
            dst = dst/self.name
        if recursive:
            for src in dfs(self, Path.get_son_iter, lambda p: p.is_file() or p.is_empty_dir()):
                src:Path
                if update and dst.exists():
                    dst.remove()
                shutil.move(src.str, (dst/src.rel_to(self)).str)
        else:
            if update and dst.exists():
                dst.remove()
            shutil.move(self.str, dst.str)


    def make_archive(self, dst, format = None):
        dst = Path(dst)
        if format == None:
            format = dst.detect_format_by_suffix()
            if format == None:
                format = 'zip'
        a = self.absolute().to_str()
        b = dst.absolute().to_str()
        shutil.make_archive(b, format, a)

    def unpack_archive_to(self, dst, format = None):
        if format == None:
            format = self.detect_format_by_suffix()
        a = self.absolute().to_str()
        b = Path(dst).absolute().to_str()
        shutil.unpack_archive(a,b,format)

    def detect_format_by_suffix(self):
        m = {
            'zip' : 'zip',
            'tar' : 'tar',
            'gz' : 'gztar',
            'bz' : 'bztar',
            'xz' : 'xztar'
        }
        ext = self.ext
        if ext:
            format = m.get(ext)
            if format:
                return format
            else:
                return ext
        
    def unzip(self, dst, print=lambda *args:None, show_state=lambda **kwargs:None):
        '''
        show_state(description='', size='', total_size='', total_compress_size='')
        '''
        dst = Path(dst).absolute().to_str()
        zf = ZipFile(self.to_str())
        l = zf.infolist()
        # file_num = len(l)
        # total_size = sum(f.file_size for f in l)
        # total_compress_size = sum(f.compress_size for f in l)
        print(f'Unzip: {self.absolute().to_str()}')
        print(f'Unzip to: {dst}')
        for i,f in enumerate(l):
            zf.extract(f, dst)
            show_state(desc=f'{i}')
            show_state(size=f.compress_size)

    def get_unique_path(self):
        '''
        return a path which is different from its siblings.
        '''
        ext = self.ext
        stem_list = [x.stem for x in self.siblings if x.ext == ext]
        stem = get_unique_name(self.stem, stem_list)
        name = f'{stem}.{ext}'
        return self.prnt/name

    def make_temp_dir(self):
        if self.is_file():
            dir_ = self.prnt
        else:
            dir_ = self
        uqtmp = (dir_/'temp').get_unique_path()
        uqtmp.mkdir()
        return uqtmp

    def remove_temp_sons(self):
        assert(self.is_dir())
        for son in self.sons:
            if son.name.startswith('temp'):
                son.remove()

    def rename_inplace(self, name:str):
        new_path = self.prnt/name
        self.rename(new_path)

    def move_all_sons_to(self, dst):
        for son in self.sons:
            son.move_to(dst)

    def move_all_out(self):
        assert(self.is_dir())
        uq_name = get_unique_name(self.name,[x.name for x in self.son_iter] + [x.name for x in self.siblings])
        if uq_name!=self.name:
            self.rename_inplace(uq_name)
            self = self.prnt/uq_name
        self.move_all_sons_to(self.prnt)
        self.remove()

del _Path



import os
from collections.abc import Iterable
from random import random
	
def cls():
	os.system('cls')
	
def is_iterable(x):
    attrs = ['__next__', '__iter__']
    if any(map(lambda a: hasattr(x,a), attrs)):
        return True

def make_plain_impl(x, ty):
    if not is_iterable(x):return [x]
    t = []
    for x0 in x:
        ty0 = type(x0)
        if ty0 == ty:
            t+=make_plain_impl(x0,ty)
        elif is_iterable(x0):
            t.append(make_plain(x0,ty))
        else:
            t.append(x0)
    return t

def make_plain(x, ty):
    if not is_iterable(x):return x
    tyx = type(x)
    if tyx != ty: return tyx(map(lambda x: make_plain(x,ty), x))
    ret = []
    for x0 in x:
        if type(x0) == ty:
            ret+=make_plain_impl(x0,ty)
        else:
            ret.append(x0)
    return tyx(ret)

def make_plain_self_impl(x):
    if not is_iterable(x):return [x]
    tyx = type(x)
    ret = []
    for x0 in x:
        if type(x0) == tyx:
            ret+=make_plain_self_impl(x0)
        elif is_iterable(x0):
            ret.append(make_plain_self(x0))
        else:
            ret.append(x0)
    return ret

def make_plain_self(x):
    if not is_iterable(x):return x
    tyx = type(x)
    ret = []
    for x0 in x:
        if type(x0) == tyx:
            ret+=make_plain_self_impl(x0)
        elif is_iterable(x0):
            ret.append(make_plain_self(x0))
        else:
            ret.append(x0)
    return tyx(ret)


def bisect_right(a, x, lo=0, hi=None, key=None):
    if hi == None:
        hi = len(a)
    while lo < hi:
        mid = (lo+hi)//2
        midv = a[mid]
        if x < key(midv):
            hi = mid
        else:
            lo = mid + 1
    return lo


def bisect_left(a, x, lo=0, hi=None, key=None):
    if hi == None:
        hi = len(a)
    while lo < hi:
        mid = (lo+hi)//2
        midv = a[mid]
        if key(midv) < x:
            lo = mid+1
        else:
            hi = mid
    return lo

def is_empty(x):
    return len(x) == 0

import heapq
from more_itertools import first as get_one

def get_min_gap_num(l:list, min_val:int = 1):
    if not len(l):
        return min_val
    heapq.heapify(l)
    v = heapq.heappop(l)
    if v>min_val:
        return min_val
    while len(l):
        v+=1
        u = heapq.heappop(l)
        if u != v:
            return v
    return v+1

def get_unique_name(name:str, src):
    if name not in src:
        return name
    d_nums = []
    for x in src:
        if f'{name}_' == f'{x[0:len(name)]}_':
            try: 
                d_num = int(x[len(name)+1: ])
            except ValueError:
                continue
            d_nums.append(d_num)
    uq_num = get_min_gap_num(d_nums,1)
    return f'{name}_{uq_num}'


__all__ = ['DictObjectAdaptor']

class DictObjectAdaptor:
    def __init__(self, o: object, keys = None):
        self.o = o
        if keys is None:
            self._keys = None
        else:
            self._keys = set(keys)

    def keys(self):
        if self._keys is None:
            return dir(self.o)
        else:
            return self._keys

    def items(self):
        for k in self.keys():
            yield k, getattr(self.o, k)

    def __getitem__(self, key):
        if self._keys is not None and key not in self._keys:
            raise KeyError
        return getattr(self.o, key)
            
    def __setitem__(self, key, value):
        if self._keys is not None:
            self._keys.add(key)
        setattr(self.o, key, value)

    def __delitem__(self, key):
        if self._keys is not None:
            del self._keys[key]
        delattr(self.o, key)

    def __contains__(self, key):
        if self._keys is not None:
            return key in self.keys
        else:
            return hasattr(self.o, key)

    def update(self, d: dict, cond=None):
        if cond is not None:
            for k, v in d.items():
                if cond(k, v):
                    self[k] = v
        for k, v in d.items():
            self[k] = v

    def get(self, key, default = None):
        if key in self:
            return self[key]
        else:
            return default



# from tqdm import tqdm
# from functools import partial

# class tqdmc(tqdm):
#     def __iter__(self):
#         def it_():
#            yield from iter(super())
#            self.close()
#         return it_() 
    

# file_proc_bar = partial(tqdm, unit = 'b', unit_scale = True)


# __all__ = ['file_proc_bar', 'tqdmc']


from functools import partial

g_str_len = 0

def compute_length(*args,**kwargs):
    sep = kwargs.get('sep')
    end = kwargs.get('end')
    arg_num = len(args)
    length = 0
    if sep: length+= max(arg_num-1,0)*len(sep)
    if end: length+=len(end)
    length+=sum(map(len,map(str,args)))
    return length

printa = partial(print, end='')

def printr(*args, len=0, **kwargs):
    printa('\r' + ' '*len)
    printa('\r')
    printa(*args, **kwargs)

class InlinePrinter:
    def __init__(self):
        self.is_first = True
        self.cur_len = 0
        
    def get_print_func(self):
        return self.print,self.printa,self.printr
    
    def print(self, *args,**kwargs):
        print(*args,**kwargs)
        self.cur_len = compute_length(*args, **kwargs)

    def printa(self, *args, **kwargs):
        printa(*args,**kwargs)
        self.cur_len = compute_length(*args, **kwargs)

    def printr(self, *args, **kwargs):
        printr(*args, len=self.cur_len,**kwargs)
        self.cur_len = compute_length(*args, **kwargs)


ginline_printer = InlinePrinter()
gprint, gprinta, gprintr = ginline_printer.get_print_func()


__all__ = ['printa', 'printr', 'InlinePrinter', 'gprint', 'gprinta', 'gprintr']


import logging

from setuptools import *


setup(
    name='cbutil',
    version='1.2.7',
    packages=find_namespace_packages('inc'),
    package_dir={'cbutil': 'inc/cbutil'},
    install_requires=[
        'base58',
        'more_itertools'
        ],
    python_requires='>=3.8',

    url='https://github.com/happyxianyu/cbutil-python',
    license='Apache License 2.0',
    author='happyxianyu',
    author_email=' happyxianyu623@outlook.com',
    description='Utility Library'
)




from collections.abc import Iterable
from collections import deque

from .util import is_iterable



#begin impl

def dfs1(n, leaf_only=True):
    '''
    example:
    list(dfs1([1,[2,3],4,[5],6])) == [1, 2, 3, 4, 5, 6]
    '''
    if(not is_iterable(n)):
        yield n
    else:
        if not leaf_only:
            yield n
        for n_ in n:
            yield from dfs1(n_, leaf_only)


def dfs2(n, is_leaf, leaf_only=True):
    '''
    exmaple:
    list(dfs2(['01', ['02','03'], '04', ['05'], '06'], lambda n: type(n) == str)) 
    == ['01', '02', '03', '04', '05', '06']
    '''
    if is_leaf(n):
        yield n
    else:
        if not leaf_only:
            yield n
        for n_ in n:
            yield from dfs2(n_, is_leaf, leaf_only)

def dfs3(n, get_sons, leaf_only=True):
    '''
    example:
    dfs([1, [2, 3], 4, [5], 6], reversed)) == [6, 5, 4, 3, 2, 1]
    '''
    if(not is_iterable(n)):
        yield n
    else:
        if not leaf_only:
            yield n
        for n_ in get_sons(n):
            yield from dfs3(n_, get_sons, leaf_only)


def dfs4(n, get_sons, is_leaf, leaf_only=True):
    '''
    example:
    (list(dfs(['01', ['02', '03'], '04', ['05'], '06'], reversed, lambda n: type(n) == str)) 
    == ['06', '05', '04', '03', '02', '01']
    '''
    if(is_leaf(n)):
        yield n
    else:
        if not leaf_only:
            yield n
        for n_ in get_sons(n):
            yield from dfs4(n_, get_sons, is_leaf, leaf_only)



def bfs1(n):
    q = deque()
    q.append(n)

    while len(q):
        for n_ in q.popleft():
            if is_iterable(n_):
                q.append(n_)
            else:
                yield n_

def bfs2(n, is_leaf):
    q = deque()
    q.append(n)

    while len(q):
        for n_ in q.popleft():
            if is_leaf(n_):
                yield n_
            else:
                q.append(n_)


def bfs3(n, get_sons):
    q = deque()
    q.append(n)

    while len(q):
        for n_ in get_sons(q.popleft()):
            if is_iterable(n_):
                q.append(n_)
            else:
                yield n_


def bfs4(n, get_sons, is_leaf):
    q = deque()
    q.append(n)

    while len(q):
        for n_ in get_sons(q.popleft()):
            if is_leaf(n_):
                yield n_
            else:
                q.append(n_)
#end impl


def dfs(n, get_sons = None, is_leaf = None, leaf_only=True):
    '''
    depth first search by preorder
    example:
    (list(dfs(['01', ['02', '03'], '04', ['05'], '06'], reversed, lambda n: type(n) == str)) 
    == ['06', '05', '04', '03', '02', '01']
    '''
    if get_sons is None and is_leaf is None:
        yield from dfs1(n,leaf_only)
    elif get_sons is None:
        yield from dfs2(n, is_leaf,leaf_only)
    elif is_leaf is None:
        yield from dfs3(n, get_sons,leaf_only)
    else:
        yield from dfs4(n, get_sons, is_leaf,leaf_only)



def bfs(n, get_sons = None, is_leaf = None):
    if get_sons == None and is_leaf == None:
        yield from bfs1(n)
    elif get_sons == None:
        yield from bfs2(n, is_leaf)
    elif is_leaf == None:
        yield from bfs3(n, get_sons)
    else:
        yield from bfs4(n, get_sons, is_leaf)




class TreeIter:
    '''
    search with record of path
    '''
    def __init__(self, root, get_sons = iter, is_leaf = lambda x: not is_iterable(x)):
        self.path = [root]
        self.get_sons = get_sons
        self.is_leaf = is_leaf

    @property
    def root(self):
        '''firt node of path'''
        return self.path[0]

    @root.setter
    def root(self,x):
        self.path[0] = x

    @property
    def cur(self):
        '''last node of path'''
        return self.path[-1]

    @cur.setter
    def cur(self,x):
        self.path[-1] = x

    @property
    def deepth(self):
        '''deepth of current path'''
        return len(self.path) -1


    def dfs(self):
        '''dfs preorder'''
        cur = self.cur
        is_leaf = self.is_leaf
        yield cur
        if is_leaf(cur):
            return

        path = self.path
        get_sons = self.get_sons
        path.append(None)
        for n in get_sons(cur):
            path[-1] = n
            yield from self.dfs()
        path.pop()

    def bfs(self):
        cur = self.cur
        is_leaf = self.is_leaf
        get_sons = self.get_sons
        path = self.path
        yield cur
        if is_leaf(cur):
            return

        q = deque()
        q.append(cur)
        prnt_num = 1
        sons_num = 0
        path.append(None)
        while len(q):
            for n_ in get_sons(q.popleft()):
                sons_num+=1
                cur = n_
                yield n_
                if is_leaf(n_):
                    continue
                else:
                    q.append(n_)
            prnt_num-=1
            if prnt_num == 0:
                # next level
                prnt_num = sons_num
                sons_num = 0
                path.append(None)
        self.path = path[0:]
            


__all__ = ['dfs', 'bfs','TreeIter']


import asyncio
import atexit
import weakref

from .coro_queue import CoroDeque

__all__ = ['CoroExecutor']

_workers_ref = set()


@atexit.register
def _cleanup():
    for ref in list(_workers_ref):  # 避免changed size
        worker = ref()
        if worker is not None:
            worker.stop_b()

class CoroExecutor:
    """
    A coroutine executor

    suffix meaning:
    a: async version
    b: block version
    nw: no wait
    """
    def __init__(self, workers_num=3):
        self._q = CoroDeque()
        self.workers = [self.Worker(self._q) for _ in range(workers_num)]

    class Worker:
        def __init__(self, q):
            self.stop_flag = False
            self.stop_event = asyncio.Event()
            self.waiting_flag = False
            self.running_flag = False
            self.q: CoroDeque = q
            self._in = []

            loop = asyncio.get_event_loop()
            self.task: asyncio.Task = loop.create_task(self.work())
            ref = weakref.ref(self)
            self._ref = ref
            _workers_ref.add(ref)

        async def work(self):
            q = self.q
            while not self.stop_flag:
                stop_event = self.stop_event

                # 等待 stop event 或 任务队列，如果 stop event 返回则直接终止该程序
                self.waiting_flag = True
                event = asyncio.Event()

                stop_future = asyncio.ensure_future(stop_event.wait())
                pop_future = asyncio.ensure_future(q.popout())
                stop_future.add_done_callback(lambda _:event.set())
                pop_future.add_done_callback(lambda _:event.set())

                await event.wait()
                if stop_future.done():
                    pop_future.cancel()
                    self.waiting_flag = False
                    return

                term = pop_future.result()
                #
                # for coro in asyncio.as_completed([stop_event.wait(), q.popout()]):
                #     term = await coro
                #     self.waiting_flag = False
                #     if stop_event.is_set():
                #         return
                #     break

                if term is CoroDeque.NoWait:
                    return
                task, future = term

                try:
                    self.running_flag = True
                    result = await task
                except Exception as e:
                    future.set_exception(e)
                else:
                    future.set_result(result)
                finally:
                    self.running_flag = False
                    await asyncio.sleep(0)  # 转移控

        async def stop(self):
            if self.stopped():
                return

            self.stop_flag = True
            self.stop_event.set()
            await self.task
            self.task = None

        def stop_b(self):
            loop = asyncio.get_event_loop()

            if loop.is_running():
                loop.create_task(self.stop())
            else:
                loop.run_until_complete(self.stop())
            self.task = None

        def terminate(self):
            if self.stopped():
                return
            if self.waiting_flag is True:
                self.stop_b()
                return
            self.task.cancel()
            self.stop_b()
            self.task = None

        def stopped(self):
            return self.task is None

        def __del__(self):
            self.stop_b()
            _workers_ref.remove(self._ref)

    async def submit(self, task, front = False) -> asyncio.Future:
        loop = asyncio.get_event_loop()
        future = loop.create_future()
        term = task, future
        if front:
            await self._q.appendleft(term)
        else:
            await self._q.put(term)
        return future

    def submit_nw(self, task, front = False):
        loop = asyncio.get_event_loop()
        future = loop.create_future()
        term = task, future
        if front:
            self._q.appendleft_nw(term)
        else:
            self._q.put_nw(term)
        return future

    async def run(self):
        while True:
            await self._q.wait_pop()

    async def wait_done(self):
        await self._q.wait_done()

    @property
    def wokers_num(self):
        return len(self.workers)

    def shutdown(self):
        for worker in self.workers:
            worker.terminate()
        self.workers.clear()

    async def close(self):
        workers = self.workers
        for worker in self.workers:
            await worker.stop()
        self.workers.clear()

    def close_b(self):
        loop = asyncio.get_event_loop()
        if loop.is_running():
            loop.create_task(self.close())
        else:
            loop.run_until_complete(self.close())
        self.workers.clear()

    def closed(self):
        return len(self.workers) == 0
